use std::{error::Error, sync::Arc};

use vulkano::{
    buffer::{IndexBuffer, Subbuffer},
    command_buffer::{AutoCommandBufferBuilder, PrimaryAutoCommandBuffer},
    descriptor_set::allocator::StandardDescriptorSetAllocator,
    pipeline::GraphicsPipeline,
};

pub trait AsRecordable<T, V, MI, I>
where
    MI: Into<IndexBuffer> + Clone,
{
    fn record_bind_commands(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        descriptor_set_allocator: &Arc<StandardDescriptorSetAllocator>,
        pipeline: &Arc<GraphicsPipeline>,
        mesh: &impl AsRenderableMesh<V, MI>,
        instances: &impl AsRenderableMeshInstance<I>,
        extra_data: &T,
    ) -> Result<(), Box<dyn Error>>;

    fn record_draw_commands(
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
        mesh: &impl AsRenderableMesh<V, MI>,
        instances: &impl AsRenderableMeshInstance<I>,
    ) -> Result<(), Box<dyn Error>> {
        match mesh.index_buffer() {
            Some(index_buffer) => {
                builder.bind_index_buffer(index_buffer.clone())?;
                unsafe {
                    builder.draw_indexed(
                        mesh.index_count(),
                        instances.instance_count(),
                        mesh.first_index(),
                        mesh.vertex_offset(),
                        instances.first_instance(),
                    )?;
                }
            }
            None => unsafe {
                builder.draw(
                    mesh.vertex_count(),
                    instances.instance_count(),
                    mesh.first_vertex(),
                    instances.first_instance(),
                )?;
            },
        }

        Ok(())
    }
}

pub trait AsRenderableMesh<V, I>
where
    I: Into<IndexBuffer> + Clone,
{
    fn vertex_buffer(&self) -> &Subbuffer<[V]>;

    fn vertex_count(&self) -> u32;

    fn first_vertex(&self) -> u32 {
        0
    }

    fn vertex_offset(&self) -> i32 {
        0
    }

    fn index_buffer(&self) -> Option<&I> {
        None
    }

    fn index_count(&self) -> u32 {
        0
    }

    fn first_index(&self) -> u32 {
        0
    }
}

pub trait AsRenderableMeshInstance<T> {
    fn instance_buffer(&self) -> &Subbuffer<[T]>;

    fn instance_count(&self) -> u32;

    fn first_instance(&self) -> u32 {
        0
    }
}
